/*
 * access.c 
 * NNTPSwitch read and parse access.conf
 *
 * $Id: access.c,v 1.23 2004/11/25 09:19:50 tommy Exp $
 */

#include "aconfig.h"
#include "nntpd.h"
#include "memory.h"
#include "aprotos.h"

int linenr;
char acfile[128];

AUTH    *auths;
PROFILE *profiles;
ALIAS   *aliases;
WILDMAT *wildmats[MAX_WILDMATS];
ACL	*acls[MAX_ACLS];

/*
 * Loading functions.
 *
 * Load the access.conf and parse each recognized statement
 */

PROTO void load_access_conf(int reload) {
    if ( reload == 0 ) {
        if ( (profiles=memmap(sizeof(PROFILE) * MAX_PROFILES)) == NULL )
	    die("Cant mmap profiles");
        memset(profiles, 0, sizeof(PROFILE) * MAX_PROFILES);
    }

    if ( (auths=memmap(sizeof(AUTH) * MAX_AUTHS)) == NULL ) {
	die("Cant mmap auths");
    }

    if ( (aliases=memmap(sizeof(ALIAS) * MAX_ALIASES)) == NULL )
	die("Cant mmap aliases");
    memset(aliases, 0, sizeof(ALIAS) * MAX_ALIASES);

    strcpy(acfile, cfg.AccessFile);
    read_acs_file(reload);

    /* do some config checks */
    if ( getprofile("default") == NULL )
	die("Profile \"default\" not defined");
    if ( getalias(DEFAULT_AUTH) == NULL )
	die("Alias \"%s\" not defined", DEFAULT_AUTH);
    if ( getalias(DEFAULT_ACCT) == NULL )
	die("Alias \"%s\" not defined", DEFAULT_ACCT);

    info("Loaded access.conf");
    info("Loaded %d Wildmats, %d Aliases, %d ACLs, %d Profiles, %d Authenticators", 
                master->numwildmats, 
                master->numaliases, 
                master->numacls, 
                master->numprofiles, 
                master->numauths);
}

/*
 * Configuration parts parser functions
 */

PROTO void read_acs_file(int reload) {
    FILE *f;
    char buf[MAX_CFGLINE], b2[MAX_CFGLINE];

    if ( (f=fopen(acfile,"r")) == NULL )
	die("Can't open profile config %s", acfile);

    linenr=0;
    while( fgets(buf, MAX_CFGLINE, f) ) {
	chop(buf);
	linenr++;

	if ( buf[0] != '\r' && buf[0] != '\n' && buf[0] != ';' && buf[0] != '#' && buf[0] != 0 ) 
        {
            if ( strncasecmp(buf, "profile", 7) == 0 ) {
		parsecfg_profile(buf, f, reload);
	    }else if ( strncasecmp(buf, "wildmat", 7) == 0 ) {
		parsecfg_wildmat(buf,f);
	    }else if ( strncasecmp(buf, "alias", 5) == 0 ) {
		parsecfg_alias(buf,f);
	    }else if ( strncasecmp(buf, "acl", 3) == 0 ) {
		parsecfg_acl(buf,f);
	    }else if ( strncasecmp(buf, "auth", 4) == 0 ) {
		parsecfg_auth(buf,f);
	    }else if ( strncasecmp(buf, "include", 7) == 0 ) {
                sscanf(buf, "%s %s", b2, acfile);
                check_path(acfile);
		read_acs_file(reload);
	    }else {
		die("Syntax error in %s on line %d", acfile, linenr);
	    }
	}
    }

    fclose(f);
}


PROTO void profile_copycfg(char *buf, FILE *f, PROFILE *pf) {
    char key[256], val[256];
    int i=0;

    while( i == 0 && fgets(buf, MAX_CFGLINE, f) ) {
	chop(buf);

	linenr++;
	if ( buf[0] == ';' || buf[0] == '#' || buf[0] == 0 )
	    continue;

	if ( strncasecmp(buf, "end", 3) == 0 ) {
	    i++;
	    break;
	}

	key[0] = 0;
	val[0] = 0;

	if ( sscanf(buf, "%*[ \t]%64s%*[ \t]%192[^\r\n]%*[\t\r\n]", key, val) != 2 )
	    die("Error on line %d of %s (%s - %s)", linenr, acfile, key, val);

	if ( strcasecmp(key, "ReadPat") == 0 ) {
            if ( getwildmat(val) == NULL )
                die("Unknown wildmat on line %d of %s", linenr, acfile);
	    strcpy(pf->ReadPat, val);
	}else if ( strcasecmp(key, "PostPat") == 0 ) {
            if ( getwildmat(val) == NULL )
                die("Unknown wildmat on line %d of %s", linenr, acfile);
	    strcpy(pf->PostPat, val);
	}else if ( strcasecmp(key, "Banner") == 0 ) {
	    strcpy(pf->Banner, val);
	}else if ( strcasecmp(key, "ReplaceBanner") == 0 ) {
	    pf->ReplaceBanner = atoi(val);
	}else if ( strcasecmp(key, "MaxConnections") == 0 ) {
	    pf->MaxConnections = atoi(val);
	}else if ( strcasecmp(key, "MaxUserConcurrent") == 0 ) {
	    pf->MaxUserConcurrent = atoi(val);
	}else if ( strcasecmp(key, "MaxHostConcurrent") == 0 ) {
	    pf->MaxHostConcurrent = atoi(val);
	}else if ( strcasecmp(key, "MaxUsers") == 0 ) {
	    pf->MaxUsers = atoi(val);
	}else if ( strcasecmp(key, "MaxSessionTime") == 0 ) {
	    pf->MaxSessionTime = atoi(val);
	}else if ( strcasecmp(key, "MaxSessionBytes") == 0 ) {
	    pf->MaxSessionBytes = atoll(val);

	}else if ( strcasecmp(key, "MaxUserBPS") == 0 ) {
	    pf->MaxUserBPS = atoi(val);
	}else if ( strcasecmp(key, "MaxUserKbit") == 0 ) {
	    pf->MaxUserBPS = atoi(val) * 1000 / 8;
	}else if ( strcasecmp(key, "MaxUserMbit") == 0 ) {
	    pf->MaxUserBPS = atoi(val) * 1000000 / 8;

	}else if ( strcasecmp(key, "MaxProfileBPS") == 0 ) {
	    pf->MaxProfileBPS = atoi(val);
	}else if ( strcasecmp(key, "MaxProfileKbit") == 0 ) {
	    pf->MaxProfileBPS = atoi(val) * 1000 / 8;
	}else if ( strcasecmp(key, "MaxProfileMbit") == 0 ) {
	    pf->MaxProfileBPS = atoi(val) * 1000000 / 8;

	}else if ( strcasecmp(key, "XComplaintsTo") == 0 ) {
	    strcpy(pf->XComplaintsTo, val);
	}else if ( strcasecmp(key, "Organization") == 0 ) {
	    strcpy(pf->Organization, val);
	}else if ( strcasecmp(key, "ForceOrganization") == 0 ) {
	    pf->ForceOrganization = atoi(val);
	}else if ( strcasecmp(key, "AddNNTPPostingHost") == 0 ) {
	    pf ->AddNNTPPostingHost = atoi(val);
	}else if ( strcasecmp(key, "NoControl") == 0 ) {
	    pf->NoControl = atoi(val);
	}else if ( strcasecmp(key, "ArticleSearch") == 0 ) {
	    pf->ArticleSearch = atoi(val);
	}else if ( strcasecmp(key, "Hostname") == 0 ) {
	    strcpy(pf->Hostname, val);
	}else if ( strcasecmp(key, "FooterFile") == 0 ) {
	    strcpy(pf->FooterFile, val);
	}else{
	    die("Unknown profile option: %s %s", key, val);
	}
    }

    pf->rl_sleep = RL_INITVAL;
    if ( pf->Hostname[0] == 0 ) 
            die("Hostname not defined for profile %s", pf->name);

    cfg_load_footer(pf);
}


/*
* If FooterFile is specified in the profile configuration
* read it here and put it into Footer[] for later use
* in the post() command.
*/
PROTO void cfg_load_footer(PROFILE *pf) 
{
    FILE *ff;
    char buf[MAX_CFGLINE];
    char tmp[MAX_FOOTER];
    char *ptr;

    if ( pf->FooterFile[0] != 0 )
    {
        pf->Footer[0] = 0;
        if ( (ff=fopen(pf->FooterFile, "r")) != NULL )
        {
            while( fgets(buf, MAX_CFGLINE, ff) )
            {
                ptr = striprn(buf);
                strcpy(tmp, pf->Footer);
                snprintf(pf->Footer, MAX_FOOTER, "%s%s\r\n", tmp, ptr);
                free(ptr);
            }

            fclose(ff);
            pf->FooterSize = strlen(pf->Footer);
        }else{
            die("Can't open FooterFile %s for Profile %s, %m", pf->FooterFile, pf->name);
        }
    }
}


/*
 * Independent configuration parsers
 */

PROTO void parsecfg_profile(char *buf, FILE *f, int reload) {
    char key[256], val[256];
    PROFILE *defp;

    key[0] = 0;
    val[0] = 0;

    if ( (sscanf(buf, "%s %s", key, val)) != 2 )
	die("Profile Syntax error on line %d of %s", linenr, acfile);

    if ( reload == 1 ) {
	if ( (defp = getprofile(val)) != NULL ) {
	    profile_copycfg(buf, f, defp);
	    return;
	}
    }

    if ( strcmp(val, "default") != 0 ) {
        /* we inherit (copy) from default profile */
	if ( (defp = getprofile("default")) == NULL ) 
	    die("Could not retrieve default profile, the default profile must be listed first.");

	memcpy((profiles+master->numprofiles), defp, sizeof(PROFILE));
    }
    strcpy((profiles+master->numprofiles)->name, val);

    profile_copycfg(buf, f, (profiles+master->numprofiles));
    master->numprofiles++;

    if ( master->numprofiles > MAX_PROFILES ) 
        die("Too many profiles (%d) increase MAX_PROFILES in nntpd.h", master->numprofiles);
}

PROTO void parsecfg_wildmat(char *buf, FILE *f) {
    char wmname[128], wmval[512];

    wmname[0] = 0;
    wmval[0] = 0;

    if ( sscanf(buf, "wildmat%*[ \t]%128[^ \t]%*[ \t]%512s", wmname, wmval) != 2 ) 
	die("Wildmat config syntax error on line %d of %s", linenr, acfile);

    if ( (wildmats[master->numwildmats] = calloc(1,sizeof(WILDMAT)) ) == NULL )
	die("Can't allocate memory for WILDMAT");

    wildmats[master->numwildmats]->name    = strdup(wmname);
    wildmats[master->numwildmats]->pattern = strdup(wmval);

    master->numwildmats++;
    if ( master->numwildmats > MAX_WILDMATS ) 
	die("Too many Wildmat entries (%d) increase MAX_WILDMATS in nntpd.h", master->numwildmats);
}


PROTO void parsecfg_alias(char *buf, FILE *f) {
    char name[MAX_STRING], lib[MAX_STRING], args[MAX_STRING];

    name[0] = 0;
    lib[0] = 0;
    args[0] = 0;

    if ( sscanf(buf, "alias%*[ \t]%256[^ \t]%*[ \t]%256[^ \t]%*[ \t]%256[^\r\n]", name, lib, args) < 2 ) 
	die("Alias config syntax error on line %d of %s", linenr, acfile);

    strcpy( (aliases+master->numaliases)->name, name);
    strcpy( (aliases+master->numaliases)->lib,  lib);
    strcpy( (aliases+master->numaliases)->args, args);
 
    master->numaliases++;
    if ( master->numaliases > MAX_ALIASES ) 
	die("Too many aliases (%d) increase MAX_ALIASES in nntpd.h", master->numaliases);
}


PROTO int parsecfg_acl(char *buf, FILE *file) {
    char ip[16], pfx[3], acs[64], profile[MAX_NAME];
    struct in_addr inaddr;
    ulong inaddrt;
    ulong ipaddrs;
    char *p;

    /*
    * Read and store ip ranges for access control
    */
    profile[0] = 0;
    if ( sscanf(buf, "acl%*[ \t]%15[0-9.]%*[/]%2[0-9] %64s%*[ \t]%32s", ip, pfx, acs, profile ) < 3 )
        die("Wrong ACL entry \"%s\" on line %d of %s", buf, linenr, acfile);

    inet_aton(ip, &inaddr);

    inaddrt = ntohl(inaddr.s_addr);
    if ( atol(pfx) == 0 )
        ipaddrs = 0xFFFFFFFF;
    else
        ipaddrs = (1 << (32-atol(pfx)))-1;

    if ( (acls[master->numacls] = calloc(1,sizeof(ACL)) ) == NULL ) 
        die("Can't allocate memory for ACL");

    if ( ! nullstr(profile) && getprofile(profile) == NULL ) 
        die("Undefined profile on line %d of %s", linenr, acfile);

    acls[master->numacls]->profile = nullstr(profile) ? getprofile("default") : getprofile(profile);

    p = strtok(acs,",");
    do {
        if ( !strcasecmp(p,"read") ) 
            acls[master->numacls]->read = 1;
        else 
        if ( ! strcasecmp(p,"stats") ) 
            acls[master->numacls]->stats = 1;
        else 
        if ( ! strcasecmp(p,"post") ) 
	    acls[master->numacls]->post = 1;
        else 
        if ( ! strcasecmp(p,"apost") ) 
            acls[master->numacls]->apost = 1;
        else 
        if ( ! strcasecmp(p,"deny") )
	    acls[master->numacls]->deny = 1;
        else 
        if ( ! strcasecmp(p,"auth") )
	    acls[master->numacls]->auth = 1;
        else 
        if ( ! strcasecmp(p,"unlimit") )
	    acls[master->numacls]->unlimit = 1;
        else {
            die("ACL Syntax Error \"%s\" on line %d of %s", p, linenr, acfile);
        }
    }while( (p=strtok(NULL,",")) );

    acls[master->numacls]->start = inaddrt;
    acls[master->numacls]->end   = inaddrt + ipaddrs;
    acls[master->numacls]->size  = ipaddrs;

    master->numacls++;
    if ( master->numacls > MAX_ACLS ) 
        die("Too many access entries (%d) increase MAX_ACLS in nntpd.h");

    return 0;
}

PROTO void parsecfg_auth(char *buf, FILE *file) {
    char b1[MAX_STRING], b2[MAX_NAME], b3[MAX_NAME], b4[MAX_NAME];
    int linenr=0;

    b1[0] = 0; b2[0] = 0; b3[0] = 0; b4[0] = 0;

    if ( sscanf(buf, "auth%*[ \t]%255s %31s %31s %31s", b1, b2, b3, b4) > 0 ) {
        if ( ( nullstr(b2) == 0 && getalias(b2) == NULL ) ||
                    ( nullstr(b3) == 0 && getalias(b3) == NULL ) )
            die("Undefined Alias on line %d of %s", linenr, acfile);

        if ( nullstr(b4) == 0 && getprofile(b4) == NULL ) 
            die("Undefined Profile on line %d of %s", linenr, acfile);

        (auths+master->numauths)->mask    = strdup(b1);
        (auths+master->numauths)->auth    = nullstr(b2) == 1 ? getalias(DEFAULT_AUTH) : getalias(b2);
        (auths+master->numauths)->acct    = nullstr(b3) == 1 ? getalias(DEFAULT_ACCT) : getalias(b3);
        (auths+master->numauths)->profile = nullstr(b4) == 1 ? getprofile(DEFAULT_PROFILE) : getprofile(b4);
        master->numauths++;
    }else{
        die("Parse error on line %d of %s", linenr, acfile);
    }

}

/*
 * Free and cleaning functions
 */

PROTO void freeaccess(void) {
    int i;

    for (i=0; i<master->numacls; i++) 
	free(acls[i]);
    info("Freed %d ACLs",master->numacls);
    master->numacls = 0;
}

PROTO void clear_wmat_aliases(void) {
    int i;

    /* Clear Wildmats */
    for (i=0; i<master->numwildmats; i++) {
        free(wildmats[i]->name);
	free(wildmats[i]->pattern);
	free(wildmats[i]);
    }
    info("Freed %d Wildmats", master->numwildmats);
    master->numwildmats = 0;

    /* Clear Aliases */
    memunmap(aliases, sizeof(ALIAS) * MAX_ALIASES);
    info("Freed %d Aliases", master->numaliases);
    master->numaliases = 0;

    return;
}

PROTO void clear_profiles(void) {
    memunmap(profiles, sizeof(PROFILE) * MAX_PROFILES);
    info("Freed %d Profiles", master->numprofiles);
    master->numprofiles = 0;
    return;
}

PROTO void clearauth(void) {
    memunmap(auths,   sizeof(AUTH)  * MAX_AUTHS);
    info("Freed %d Authenticators", master->numauths);
    master->numauths = 0;
    return;
}

/*
 * Reloading Functions
 */

PROTO void reload_access_conf(void) {
    clear_wmat_aliases();
    freeaccess();
    clearauth();
    load_access_conf(1);
}

/*
 * ACL Functions
 */

PROTO ACL* getacl(ulong ipaddr) {
    int i;

    for (i=0; i<master->numacls; i++) {
	if ( ipaddr >= acls[i]->start && ipaddr <= acls[i]->end )
	    return acls[i];
    }
    return NULL;
}

/*
 * Wildmat Functions
 */

PROTO const char* getwildmat(const char* name) {
    int i;

    if ( name == NULL )    return DEFAULT_WILDMAT;
    if ( name[0] == '\0' ) return DEFAULT_WILDMAT;

    for (i=0; i<master->numwildmats; i++)
        if ( ! strcmp(wildmats[i]->name, name) )
            return wildmats[i]->pattern;
    
    return NULL;
}

/*
 * Alias Functions
 */

PROTO ALIAS* getalias(const char *name) {
    int i;

    for (i=0; i<master->numaliases; i++)
        if ( strcmp((aliases+i)->name, name) == 0 )
            return (aliases+i);
    
    return NULL;
}

/*
 * Profiles Functions
 */

PROTO PROFILE* getprofile(const char *name) {
    int i;

    for (i=0; i<master->numprofiles; i++)
	if ( strcmp((profiles+i)->name, name) == 0 )
	    return (profiles+i);

    return NULL;
}

/*
 * Auth Functions
 */

PROTO AUTH* getauth(const char *username) {
    int i;

    for (i=0; i<master->numauths; i++)
	if ( uwildmat_simple(username, (auths+i)->mask) )
	    return (auths+i);

    return NULL;
}

