/*
 * NNTPSwitch client and user handling routines
 * $Id: client.c,v 1.5 2004/11/11 13:32:44 tommy Exp $
 */


#include "nntpd.h"
#include "aconfig.h"
#include "aprotos.h"


PROTO void free_client(int clientid) {
    memset(master->clients+clientid, '\0', sizeof(CLIENT));
}


/*
 * Find and allocate a client slot from the client's pool accepting
 * the concurrent connection limits. A client slot needs to be
 * reserved before we can fork() the client in it's own process.
 */

PROTO CLIENT* find_slot(ulong ip4addr, int *concur, int maxcon) {
    int i;
    int cc=0;
    CLIENT *cl;

    /* check for both a free slot and the number of   
     * concurrent sessions at once to speed things up */

    for (i=0; i<MAX_PROC; i++) {
	if ( (master->clients+i)->ip4addr == ip4addr )
	    cc++;

	if ( (master->clients+i)->inuse == 0 ) {
	    cl = (master->clients+i);   // set client pointer to one from the list
	    memset(cl, '\0', sizeof(CLIENT));
	    cl->inuse++;
	    cl->id = i;
	    cl->ip4addr = ip4addr;

	    /* check for more concurrent connections */
	    for ( ; i<MAX_PROC; i++) {
		if ( (master->clients+i)->ip4addr == ip4addr )
		    cc++;

		if ( maxcon > 0 && cc > maxcon ) {
		    (*concur) = cc;
		    cl->inuse = 0;
		    cl->ip4addr = 0;
		    return NULL;
		}
	    }
	    (*concur) = cc;
	    cl->concur = cc;
	    cl->starttime = time(NULL);
	    return cl;
	}
    }
    syslog(LOG_ERR, "Too many slots requested, increase MAX_PROC in nntpd.h");
    return NULL;
}


/*
 * Find a user id matching the key or create a new one 
 * if he isn't connected yet.
 */

PROTO USER* find_user(char *ky, PROFILE *prf) 
{
    int i, id = -1;
    USER *u;

    semlock_lock(master->semid);

    /* check if its a known user */
    for (i=0; i<MAX_USERS; i++) 
	if ( strcmp((master->users+i)->key, ky) == 0 ) 
	{
	    ((master->users)+i)->connections++;
    	    semlock_unlock(master->semid);
	    return (master->users+i);
	}

    /* find first empty user slot */
    for (i=0; i<MAX_USERS; i++)
	if ( (master->users+i)->key[0] == 0 ) 
	{
	    id = i;
	    break;
	}

    /* we're full */
    if ( id == -1 ) 
    {
        semlock_unlock(master->semid);
	return NULL;
    }

    /* initialize the slot */
    u = (master->users)+id;
    memset(u, '\0', sizeof(USER));

    strcpy(u->key, ky);
    u->connections++;
    u->id = id;
    u->bytesleft = prf->MaxSessionBytes;
    prf->numusers++;
    master->numusers++;

    semlock_unlock(master->semid);
    return u;
}

PROTO USER * change_user(USER *olduser, char *newky, PROFILE *prf) {
    USER *newuser;

    newuser = find_user(newky, prf);
    if ( newuser == olduser || newuser == NULL )
	return olduser;

    drop_user(olduser->id, prf);
    return newuser;
}

PROTO void drop_user(uint id, PROFILE *prf) {
    USER *u;

    u = (master->users+id);
    semlock_lock(master->semid);
    u->connections--;

    if ( u->connections == 0 ) {
#ifdef USE_REALUSER
        if ( u->realuser == 1 )
	    prf->realusers--;
#endif
	memset(u, '\0', sizeof(USER));
        prf->numusers--;
        master->numusers--;
    }
    semlock_unlock(master->semid);

}

