/*
 * nntpd.c  General client allocation and command parser routines
 * $Id: nntpd.c,v 1.67 2004/11/20 22:25:28 tommy Exp $
 */

#include "nntpd.h"
#include "aconfig.h"
#include "setproctitle.h"
#include "aprotos.h"

char *helptxt = "100 Supported Commands\r\n"
                "   MODE READER\r\n"
                "   AUTHINFO USER <username>\r\n"
                "   AUTHINFO PASS <password>\r\n"
                "   LIST <active|newsgroups|overview.fmt> <pattern>\r\n"
                "   XGTITLE <pattern>\r\n"
                "   GROUP <newsgroup>\r\n"
                "   LISTGROUP <newsgroup>\r\n"
                "   NEWGROUPS <yy><yymmdd> <hhmmss>\r\n"
                "   OVER <range|msgid>\r\n"
                "   XOVER <range|msgid>\r\n"
                "   XHDR <header> <range|msgid>\r\n"
                "   XPAT <header> <range|msgid> <pattern> <pattern..>\r\n"
                "   NEWNEWS <newsgroup> <yymmdd> <hhmmss> <gmt|utc>\r\n"
                "   STAT <msgid|number>\r\n"
                "   HEAD <msgid|number>\r\n"
                "   BODY <msgid|number>\r\n"
                "   ARTICLE <msgid|number>\r\n"
                "   POST\r\n"
                "   NEXT\r\n"
                "   LAST\r\n"
                "   HELP\r\n"
                "   DATE\r\n"
                "   QUIT\r\n"
                "The NEWNEWS command is only partly supported for single newsgroups only.\r\n"
                "The ARTICLE command can only be used inside a newsgroup.\r\n"
                "\r\n"
                "Chiparus NNTPSwitch-%s (c) 2002\r\n"
                ".\r\n";

CMDINFO commands[] = {       /* auth, acct, limit, usecount */
    { "MODE",       cmd_mode,       1, 0, 0, 0 },
    { "SLAVE",      cmd_slave,      0, 0, 0, 0 }, /* not in help */
    { "AUTHINFO",   cmd_authinfo,   0, 0, 0, 0 },
    { "LIST",       cmd_list,       1, 0, 0, 0 },
    { "XGTITLE",    cmd_xgtitle,    1, 0, 0, 0 },
    { "GROUP",      cmd_group,      1, 0, 0, 0 },
    { "LISTGROUP",  cmd_listgroup,  1, 0, 0, 0 },
    { "NEWGROUPS",  cmd_newgroups,  1, 0, 0, 0 },
    { "OVER",       cmd_over,       1, 0, 0, 0 },
    { "XOVER",      cmd_xover,      1, 0, 0, 0 },
    { "XHDR",       cmd_xhdr,       1, 0, 0, 0 },
    { "XPAT",       cmd_xpat,       1, 0, 0, 0 },
    { "NEWNEWS",    cmd_newnews,    1, 0, 0, 0 },
    { "STAT",       cmd_stat,       1, 0, 0, 0 },
    { "HEAD",       cmd_head,       1, 1, 1, 0 },
    { "BODY",       cmd_body,       1, 1, 1, 0 },
    { "ARTICLE",    cmd_article,    1, 1, 1, 0 },
    { "POST",       cmd_post,       1, 0, 0, 0 },
    { "NEXT",       cmd_next,       1, 0, 0, 0 },
    { "LAST",       cmd_last,       1, 0, 0, 0 },
    { "HELP",       cmd_help,       0, 0, 0, 0 },
    { "DATE",       cmd_date,       0, 0, 0, 0 },
    { "QUIT",       cmd_quit,       0, 0, 0, 0 },
    { "STATISTICS", cmd_statistics, 0, 0, 0, 0 },
    { "IHAVE",      cmd_unsupported,0, 0, 0, 0 }, /* not in help */
    { "GET",        cmd_http,       0, 0, 0, 0 }, 
    { NULL,         cmd_unsupported,0, 0, 0, 0 },
};


/* Log statistics about the leaving client */
PROTO void client_log_exit(CLIENT *cl) {
    CMDINFO *cmd;
    char buf[1024] = "\0";

    syslog(LOG_NOTICE,"%s: connection closed articles %u bytes %llu groups %u posts %u postbytes %llu",
                client->hostname, client->articles, client->bytes, 
                client->groups, client->posts, client->postbytes);

    for( cmd = commands; cmd->command; cmd++ ) 
        sprintf(buf, "%s %s %d", buf, cmd->command, cmd->usecount);

    syslog(LOG_NOTICE, "%s: commands usage%s UNKNOWN %d", cl->hostname, buf, cmd->usecount);
}


PROTO void client_error(CLIENT *client, char *txt) {
    syslog(LOG_NOTICE, "%s: client error %s ",client->hostname, txt);
    writeclient(client,MSG_GENERIC_ERROR,txt);
    close_child(client, true);
    exit(0);
}


PROTO void close_child(CLIENT *client, bool doacct) 
{
    client_log_exit(client);

    if ( doacct == true && client->authenticated ) 
    {
        if ( ! load_acct_module() ) 
	{
            dl_acct_entry(client, &cfg, client->auth->acct->args);
            unload_acct_module();
        }else{
            syslog(LOG_DEBUG,"%s: accounting vanished", client->hostname);
        }
    }

    /* FIXME: After here, syslog might be fux0red because of accounting modules */

    client->profile->connections--;
    drop_user(client->user->id, client->profile);
    free_client(client->id);
    master->connections--;

    return;
}



/* this is called directly after forking a child and after client is allocated */
PROTO void run_child(CLIENT *cl) {
    char *rc = NULL;
    struct hostent *hp = NULL;
    client = cl;

    setproctitle("nntpswitchd %u: %s initializing", client->id, inet_ntoa(client->addr.sin_addr) );

    /* 
     * Do basic initialize of the client 
     */

    client->ip4addr = ntohl(client->addr.sin_addr.s_addr);
    if ( cfg.DNSLookups == 1 ) {
        hp = gethostbyaddr((char*)&client->addr.sin_addr.s_addr, sizeof(client->addr.sin_addr.s_addr), AF_INET);
        client->hostname = strdup(hp != NULL ? (char *)hp->h_name : inet_ntoa(client->addr.sin_addr));
    }else{
        client->hostname = strdup(inet_ntoa(client->addr.sin_addr));
    }

    client->errstr          = NULL;
    client->group           = NULL;
    client->groupserver     = NULL;
    client->profile         = client->acl->profile;

    client->profile->connections++;
    master->connections++;

    if ( (client->bbuf  = malloc(cfg.BufSize+1)) == NULL ) 
        client_error(client,"No room for bbuf");
    
    if ( (client->sbuf  = malloc(cfg.BufSize+1)) == NULL ) 
        client_error(client,"No room for sbuf");

    openlog("nntpswitchd", LOG_PID, LOG_NEWS);
    syslog(LOG_NOTICE, "connect from %s: clientid %u clientnum %u total %u",
                        client->hostname, client->id, client->concur, master->connections);
    

    /*
     * Find a *USER based on the profile we're connected to. Check
     * for any connection or user limits.
     */

    client->user = find_user(client->hostname, client->profile);

    if ( client->user == NULL ) {
        sleep(cfg.RetryDelay);
        swriteclient(client, MSG_SERVERFULL);
        syslog(LOG_ERR, "Too many global users, increase MAX_USERS");
        close_child(client, true);
        return;
    }

    if ( client->acl->unlimit != 1 && 
                client->user->connections > client->profile->MaxUserConcurrent ) {
        sleep(cfg.RetryDelay);
        writeclient(client, MSG_TOOMANY_CONCURRENT, client->profile->MaxUserConcurrent);
        close_child(client, true);
        return;
    }

    if ( client->profile->connections > client->profile->MaxConnections )
    {
        sleep(cfg.RetryDelay);
        swriteclient(client, MSG_TOOMANYCONNSPRF);
        close_child(client, true);
        return;
    }

    if ( client->profile->numusers   >= client->profile->MaxUsers )
    {
        sleep(cfg.RetryDelay);
        swriteclient(client, MSG_TOOMANYUSERSPRF);
        close_child(client, true);
        return;
    }

#ifdef USE_REALUSER
    if ( client->user->realuser == 0  ) {
        client->profile->realusers++;
        client->user->realuser = 1;
    }
#endif

    init_ratelimiter(client);

    if ( client->profile->ReplaceBanner == 1 ) {
        writeclient(client, "%u %s", client->acl->post ? 200 : 201, client->profile->Banner);
    }else{
        writeclient(client,"%u %s NNTPSwitch-%s, %u groups available, %s, slot %u, connections %u\r\n", 
                         client->acl->post ? 200 : 201,   client->profile->Banner, VERSION, master->numgroups,
                         client->acl->post ? "posting allowed" : "no posting",
                         client->id + 1, client->concur
                         );
    }

    setproctitle("nntpswitchd %u: %s idle", client->id, client->hostname);

    /* main client loop */
    errno=0;

    while ( 
            client->error == 0 && 
            (rc=readclient(client,client->bbuf,cfg.BufSize,0)) != NULL && 
            time(NULL) - client->starttime < client->profile->MaxSessionTime &&
            !(client->user->bytesleft != 0 && client->user->bytesleft < 0)
           ) 

    {
        if ( parsecmd(client, client->bbuf) )
            break;
        setproctitle("nntpswitchd %u: %s idle", client->id, client->hostname);
        errno=0;
    }

    /*
     * check what kind of errors we have or had en exit the client based on that
     * this code in general and way of errorhandling is worse than lame
     */

    if ( client->error > 0 ) {
        syslog(LOG_DEBUG,"%s: timeout or error %s", client->hostname, client->errstr);

    }else if ( client->user->bytesleft != 0 && client->user->bytesleft < 0 ) {
        syslog(LOG_DEBUG,"%s: session byte limit reached", client->hostname);
        swriteclient(client, MSG_BYTELIMIT);

    }else if ( time(NULL) - client->starttime >= client->profile->MaxSessionTime ) {
        syslog(LOG_DEBUG,"%s: session time limit reached %s", client->hostname, client->timeout ? "(timeout)" : "(normal)");
        swriteclient(client, MSG_SESSIONLIMIT);

    }else if ( rc == NULL && client->timeout == 1 ) {
        syslog(LOG_DEBUG,"%s: client timeout after %d seconds", client->hostname, cfg.ClientReadTimeout);
        writeclient(client,"205 Timeout after %d seconds, articles %u, bytes %llu, groups %u, posts %u, postbytes %llu\r\n", 
                           cfg.ClientReadTimeout, client->articles, client->bytes, 
                           client->groups, client->posts, client->postbytes);

    }else if ( errno ) {
        syslog(LOG_DEBUG,"%s: timeout or error fixme %m", client->hostname);
        writeclient(client,"205 Timeout/Error articles %u, bytes %llu, groups %u, posts %u, postbytes %llu\r\n",
                            client->articles, client->bytes, 
                            client->groups, client->posts, client->postbytes);
    
    }else{
        syslog(LOG_DEBUG,"%s: eof from client", client->hostname);
    }

    client_quit(client);
    close_child(client, true);
    return;
}


/* 
 * return true to exit, false to continue accepting commands
 */
PROTO int parsecmd(CLIENT *client, char *buf) {
    char a0[512],a1[512];
    CMDINFO *cmd;
    int i;

    a0[0] = 0; a1[0] = 0;
    sscanf(buf, "%31[^\r\n\t ] %480[^\r\n]", a0, a1);
    if (a0[0] != '\0') {
        for( cmd = commands; cmd->command; cmd++ ) 
            if ( !strcasecmp(a0,cmd->command) ) 
                break;

        client->command = cmd;
        cmd->usecount++;

        if ( cmd->num == cmd_authinfo ) {
            if ( cfg.LogCommands )
                syslog(LOG_DEBUG, "%s: command %s", client->hostname, a0);
            setproctitle("nntpswitchd %u: %s command %s", client->id, client->hostname,a0);
        }else{
            if ( cfg.LogCommands )
                syslog(LOG_DEBUG, "%s: command %s %s", client->hostname, a0, a1);
            setproctitle("nntpswitchd %u: %s command %s %s", client->id, client->hostname,a0,a1);
        }

        if ( cmd->auth == 1 && client->acl->auth == 1 && client->authenticated == 0 ) {
           return swriteclient(client, MSG_AUTHREQUIRED);
        }

        switch( cmd->num ) {

            case cmd_quit:
                writeclient(client,"205 Exit articles %u, bytes %llu, groups %u, posts %u, postbytes %llu\r\n",
                                    client->articles, client->bytes, 
                                    client->groups, client->posts, client->postbytes);
                client_quit(client);
                close_child(client, true);
                _exit(0);

            case cmd_http:
                swriteclient(client, MSG_COMMAND_HTTP);
                client_quit(client);
                close_child(client, true);
                _exit(0);

            case cmd_help:
                return writeclient(client, helptxt, VERSION);

            case cmd_date:
                send_date(client);
                return 0;

            case cmd_mode:
                if ( !strcasecmp(a1,"reader") ) {
                    if ( client->connected == 1 ) {
                        if ( ! writeserver(client,"MODE READER\r\n") ) return 1;
                        if ( (readserver(client, a1, MAX_SERVERRSP-1)) == NULL ) return 1;
                        if ( atoi(a1) > 299 ) {
                            syslog(LOG_ERR,"%s: mode reader check server %s response %s",client->hostname, client->lastserver, a1);
                            writeclient(client,a1);
                            return 1;
                        }
                    }
                    return writeclient(client,"%u Sure!\r\n", client->acl->post ? 200 : 201);
                }else
                    return swriteclient(client,"500 Unsupported MODE command\r\n");

            case cmd_slave:
                return swriteclient(client,"202 Sure!\r\n");

            case cmd_authinfo:
                return cmdauthinfo(client, a1);
           
            case cmd_list:
                return cmdlist(client, a1);
            
            case cmd_xgtitle:
                return cmdxgtitle(client, a1);

            case cmd_group:
                group(client, a1, false);
                return 0;
           
            case cmd_newgroups:
                newgroups(client, a1);
                return 0;

            case cmd_listgroup:
                if ( strlen(a1) > 0 ) {     
                    // groupname has been specified by user 
                    if ( group(client, a1, true) == 0 ) {
                        return passcmd(client, a0, a1, cmd);
                    }else{
                        return swriteclient(client, MSG_NOSUCHGROUP);
                    }
                }else{
                    if ( client->group == NULL )
                        // no current group selected 
                        return swriteclient(client, MSG_NOGROUP);
                    else
                        // no group given, use current selected group 
                        return passcmd(client, a0, "", cmd); 
                }
                return 0;

            case cmd_xover:
                if ( client->group == NULL ) 
                    return swriteclient(client, MSG_NOGROUP);
                else
                    return passcmd(client,a0, a1, cmd);

            case cmd_xpat:
            case cmd_xhdr:
                if ( client->group == NULL ) {
                    return swriteclient(client, MSG_NOGROUP);
                }else{
                    if ( cfg.DenyNoIndexField ) {
                        sscanf(a1,"%128[^\r\n\t ]",buf);

                        if ( isinoverviewfmt(buf) ) {
                            return passcmd(client,a0, a1, cmd);
                        }else{
                            return swriteclient(client, MSG_NOPATINDEX);
                        }

                    }else{
                        return passcmd(client,a0, a1, cmd);
                    }

                }

            case cmd_over:
                if ( client->group == NULL ) 
                    return swriteclient(client, MSG_NOGROUP);
                else
                    return passcmd(client, "XOVER", a1, cmd);

            case cmd_head:
                if ( catch_art_server(client, a1) == false )
                    return 0;

                return passcmd(client,a0, a1, cmd);

            case cmd_article:
            case cmd_body:
                if ( catch_art_server(client, a1) == false ) 
                    return 0;

                client->articles++;
                client->serverarts++;
                client->grouparts++;
                client->profile->articles++;
                return passcmd(client,a0, a1, cmd);

            case cmd_newnews:
                sscanf(a1,"%256[^\r\n\t ] %256[^\r\n]",a0,buf);
                if ( !strcmp(a0,"newnews") || strchr(a0,'*') != NULL || strchr(a0,'!') != NULL ) {
                    return swriteclient(client, MSG_NONEWNEWS);
                }else{
                    if ( group(client, a0, true) == 0 ) {
                        return passcmd(client, "NEWNEWS", a1, cmd);
                    }else{
                        return swriteclient(client, MSG_NOSUCHGROUP);
                    }
                }
                return 0;

            case cmd_stat:
                if ( catch_art_server(client, a1) == false )
                    return 0;

                if ( ! writeserver(client,"%s %s\r\n",a0,a1) ) break;
                if ( (readserver(client, buf, 256)) == NULL ) {
                    client_error(client, UMSG_SERVERTIMEOUT);
                }else{
                    return swriteclient(client,buf);
                }
                return 0;

            case cmd_last:
            case cmd_next:
                if ( client->group == NULL ) 
                    return swriteclient(client, MSG_NOGROUP);
                else {
                    if ( ! writeserver(client,"%s %s\r\n",a0,a1) ) break;
                    if ( (readserver(client, buf, 256)) == NULL ) {
                        client_error(client, UMSG_SERVERTIMEOUT);
                    }else{
                        return swriteclient(client,buf);
                    }
                }
                return 0;

            case cmd_post:
                if ( client->acl->post == 0 )
                    return swriteclient(client, MSG_NOPOSTING);
                else
                    return post(client);

            case cmd_unsupported:
                if ( cmd->usecount >= cfg.DropUnknown ) {
                    writeclient(client, "500 Dropped after %d invalid commands", cfg.DropUnknown);
                    client_quit(client);
                    close_child(client, true);
                    _exit(0);
                }
                return swriteclient(client, MSG_UNSUPPORTED);

            case cmd_statistics:
                if ( client->acl->stats != 1 ) 
                    return swriteclient(client, MSG_UNSUPPORTED);

                writeclient(client, "100 Statistics\r\n");
                writeclient(client, "uptime %u", time(NULL) - master->serverstart);
                writeclient(client, "-- name, connections, users, realusers, bytes, arts");
                for (i=0; i<master->numprofiles; i++) {
                    writeclient(client, "profile %s %u %u %u %llu %u\r\n", 
                                    (profiles+i)->name,
                                    (profiles+i)->connections,
                                    (profiles+i)->numusers,
                                    (profiles+i)->realusers,
                                    (profiles+i)->bytes,
                                    (profiles+i)->articles);
                }
                writeclient(client, ".");
                return 0;

            default:
                return swriteclient(client, MSG_NOTIMPLEMENTED);
        }
    }
    return 0;
}

