
#include "aconfig.h"
#include "nntpd.h"
#include "aprotos.h"

#define CLEARBUF client->bbuf[0] = 0;
#define CLEARPASS memset(&client->password, 0, MAX_USERPASS);


PROTO void client_quit(CLIENT *client) {
    if ( client->connected == 1 ) {
        closegroup(client);
        disconnect_server(client);
    }
}

PROTO int send_date(CLIENT *client) {
    struct timeval tv;
    struct tm *ptm;
    char timestr[40];

    gettimeofday(&tv,NULL);
    ptm = gmtime(&tv.tv_sec);
    strftime(timestr, sizeof(timestr), "%Y%m%d%H%M%S",ptm);

    /* FIXME: This ain't no UTC! */
    return writeclient(client,"111 %s UTC\r\n",timestr);
}


PROTO int passcmd(CLIENT *client, char *command, char *args, CMDINFO *cmd) {
    int     i = 1;
    char    chkbuf[5];
    ssize_t bread = 0;
    CLEARBUF

    if ( ! writeserver(client, "%s %s\r\n", command, args) ) return -1;

    if ( (readserverline(client->bbuf, cfg.BufSize)) == NULL ) return -1;
    if ( swriteclient(client, client->bbuf) ) return -1;
    if ( atoi(client->bbuf) > 399 ) i=0;

    checkbuf_init(chkbuf);
    while (i && ! client->error) {
        if ( (bread = nreadserver(client, client->bbuf, cfg.BufSize)) == 0 )
            return -1;

        if ( checkbuf_isend(chkbuf, client->bbuf, bread) ) 
            i = 0;

        if ( slwriteclient(client, client->bbuf, bread, cmd->limit) ) 
            return -1;

        if ( cmd->acct == 1 )
            client->bytes += (ulong64)bread;

        if ( cmd->acct == 1 && client->user->bytesleft > 0 )
            client->user->bytesleft -= (ulong64)bread;

        client->groupbytes  += (ulong64)bread;
    }
    return 0;
}

PROTO void closegroup(CLIENT *client) {
    if ( client->group!= NULL && (client->grouparts > 0 || client->groupbytes > 0) ) {
        syslog(LOG_NOTICE, "%s: group %s closed articles %u bytes %llu", 
                                client->hostname, 
                                client->group->newsgroup, 
                                client->grouparts, 
                                client->groupbytes);
        client->grouparts = 0;
        client->groupbytes = 0;
    }
}

PROTO int group(CLIENT *client, const char *group, char nogroupecho) {
    char b[MAX_SERVERRSP];
    ACTIVE *newgroup;
    SERVER *svr;
    CLEARBUF

    if ( nullstr(group) ) {
        swriteclient(client,"500 You must select a group\r\n");
        return 1;
    }
  
    /* only change group if it has changed */
    if ( client->group == NULL || strcmp(client->group->newsgroup, group) != 0 ) {

        /* check for read permission */
        if ( ! match_expression(group, getwildmat(client->profile->ReadPat), 0) ) {
            if ( ! nogroupecho )
                swriteclient(client, MSG_NOSUCHGROUP);
            return 1;
        }

	/* find and load our new group */
        if ( (newgroup = getgroup(group)) == NULL ) {
            if ( ! nogroupecho ) 
                swriteclient(client, MSG_NOSUCHGROUP);
            return 1;
        }

        if ( (svr=getserver(newgroup->server)) == NULL )
            return swriteclient(client, "500 Cant find server for group\r\n");

        /* disconnect first if we have to change server */
        if ( client->groupserver != NULL && 
             client->connected == 1 &&
             strcmp(svr->Groups, client->groupserver->Groups) != 0 ) 
	{
            disconnect_server(client);
        }

	/* close group and reconnect */
        closegroup(client);
        client->group = newgroup;

    	switch ( connect_groupserver(client, svr->Name) ) {
	    case CONNECT_ERR:
            	return swriteclient(client, "500 Error selecting server\r\n");

            case CONNECT_DOWN:
                sleep(cfg.DownDelay);
                return writeclient(client, MSG_SERVER_DOWN);

            default:
                break;
        }

	/* reset by connect_groupserver (via disconnect_server FIXME) */
	client->group = newgroup;

        client->groups++;
    }

    if ( ! writeserver(client, "GROUP %s\r\n", group) ) return -1;
    if ( ! readserver(client, b, MAX_SERVERRSP-1) ) return -1;

    syslog(LOG_NOTICE, "%s: group %s", client->hostname, group);

    if ( ! nogroupecho ) return swriteclient(client,b);
    return 0;
}

PROTO int cmdlist(CLIENT *client, char *args) {
    char cmd[16];

    if ( strlen(args) == 0 )
        return listactive(client,"*");

    CLEARBUF
    sscanf(args,"%16[^\r\n\t ] %256[^\r\n]", cmd, client->bbuf);
    if ( client->bbuf[0] == 0 ) strcpy(client->bbuf,"*");

    if ( !strncasecmp(cmd,"active.times",12) ) {
        listactivetimes(client);

    }else if ( !strncasecmp(cmd,"active",6) ) {
        listactive(client,client->bbuf);

    }else if ( !strncasecmp(cmd,"newsgroups",10)) {
        listnewsgroups(client, client->bbuf);

    }else if ( !strncasecmp(cmd,"overview",8)) {
        listoverviewfmt(client);

    }else{
        return swriteclient(client,"500 Unsupported LIST command\r\n");
    }
    return 0;
}

PROTO int cmdxgtitle(CLIENT *client, char *args) {
    if ( strlen(args) == 0 )
        return listxgtitle(client,"*");

    CLEARBUF
    sscanf(args,"%256[^\r\n]", client->bbuf);
    if ( client->bbuf[0] == 0 ) strcpy(client->bbuf,"*");

    listxgtitle(client, client->bbuf);
    return 0;
}

PROTO int cmdauthinfo(CLIENT *client, char *args) {
    char subcmd[17], userpass[65];
    char *err;
    USER *newuser;
    PROFILE *oldprf;
    int change_profile;

    sscanf(args, "%16[^\r\n\t ] %64[^\r\n]", subcmd, userpass);
    
    if ( !strcasecmp(subcmd,"user") ) {
        strcpy(client->username, userpass);
    }else if ( !strcasecmp(subcmd,"pass") ) {
        strcpy(client->password, userpass);
    }else{
        return swriteclient(client,"501 Unsupported AUTHINFO command\r\n");
    }
    
    if ( client->username[0] == 0 || client->password[0] == 0 )
        return swriteclient(client, MSG_MOREAUTHREQ);
    
    client->auth = getauth(client->username);
    

    if ( ! load_auth_module() ) {
	client->authenticated = 0;
        err = strdup(dl_check_auth(client, &cfg, client->auth->auth->args));
        unload_auth_module();
        syslog(LOG_DEBUG,"authinfo result %s", err);
        CLEARPASS
        
        /* switch userid and profile */
        if ( client->authenticated ) {

            /* Check and verify we are allowed in our new profile */

	    change_profile = 0;
            oldprf = client->profile;
            if ( client->profile != client->auth->profile ) {
                if ( client->auth->profile->connections > client->auth->profile->MaxConnections ) {
                    client->authenticated--;
                    return swriteclient(client, MSG_AUTH_TOOMANYCON);
                }else
                if ( client->auth->profile->numusers > client->auth->profile->MaxUsers ) {
                    client->authenticated--;
                    return swriteclient(client, MSG_AUTH_TOOMANYUSR);
                }else{
	  	    change_profile = 1;
                }
            }

            /* Find our other connections from this user to see if we are allowed */

            newuser = find_user(client->username, client->auth->profile);
            if ( newuser != client->user && newuser != NULL ) {
                if ( newuser->connections > client->auth->profile->MaxUserConcurrent ) { 
                    client->authenticated--;
                    drop_user(newuser->id, client->profile);
                    sleep(cfg.RetryDelay);
                    return swriteclient(client, MSG_AUTH_USERFULL);
                }
            }


	    if ( change_profile == 1 ) {
                client->profile->connections--;
                client->profile = client->auth->profile;
                client->profile->connections++;
	    }


            drop_user(client->user->id, oldprf);
            client->user = newuser;
            client->user->bytesleft = client->initbytes;
            syslog(LOG_DEBUG, "%s: user logged in %s", client->hostname, client->username);

            /* init ratelimiter */
            init_ratelimiter(client);
        }
        
        return swriteclient(client, err);
    }else{
        CLEARPASS
        return swriteclient(client, MSG_AUTH_ERR);
    }
}


/*
 * Find article server based on message id. Return true
 * on success, false on error.
 */
PROTO int find_article_server(CLIENT *cl, char *msgid) {
    SERVER *svr, *olds;
    int i, oldlev = -1, x;

    if ( msgid[0] != '<' )
        return CHECK_ARTICLE_FAIL;

    olds = cl->groupserver;
    if ( cl->groupserver != NULL ) 
    {
        svr = cl->groupserver;
        oldlev = svr->Level;

        x = check_headperm(cl, msgid);
        if ( x == CHECK_ARTICLE_OK || x == CHECK_ARTICLE_NOPERM ) 
            return x;
    }

    for( i=0; i<master->numservers; i++ ) 
    {
        if ( (servers+i)->Level != oldlev && 
             (servers+i) != olds && 
             strcmp((servers+i)->Name, "default") != 0 ) 
        {
            svr = (servers+i);
            oldlev = svr->Level;

            if ( connect_groupserver(cl, svr->Name) == CONNECT_OK ) {
                x = check_headperm(cl, msgid);
                if ( x == CHECK_ARTICLE_OK || x == CHECK_ARTICLE_NOPERM ) 
                        return x;
            }
        }
    }

    /* no server has article */
    return CHECK_ARTICLE_FAIL;
}


/*
 * Connect to a backend server taking the backup and balancing
 * policies in account. Return false on some error, true on
 * success.
 */
PROTO bool connect_groupserver(CLIENT *cl, char *server) {
    SERVER *svr, *primary;

    /* disconnect first if we have to change server */
    if ( cl->connected == 1 && strcmp(cl->groupserver->Name, server) != 0 ) {
        disconnect_server(cl);
    }
    
    if ( cl->connected != 1 ) {
        /* Get the real server and take balancing stuff in account */
        if ( (svr=selectserver_primary(server)) == NULL )
            return CONNECT_ERR;

        primary = svr;  /* remember where we started with */
        cl->groupserver = svr;

        if ( connect_server(cl, svr) ) {
            if ( errno ) {
                syslog(LOG_ERR,"%s: remote server %s down %m", cl->hostname, cl->groupserver->Name);
                errno = 0;
            }
    
            /* On server down find backup servers and try to conect to them */
            while( cl->connected != 1 && (svr = selectserver_backup(cl->groupserver, primary)) != NULL ) {
                cl->groupserver = svr;
                if ( connect_server(cl, svr) ) {
                    if ( errno ) {
                        syslog(LOG_ERR,"%s: backup server %s down %m", cl->hostname, cl->groupserver->Name);
                        errno = 0;
                    }
                }
            }
        
            if ( cl->connected != 1 ) {
                client->group = NULL;
                client->groupserver = NULL;
                return CONNECT_DOWN;
            }
        }
    }

    return CONNECT_OK;
}


PROTO bool check_headperm(CLIENT *cl, char *msgid) {
    int     i = 1;
    char    chkbuf[5];
    ssize_t bread = 0;
    char    *ng, *ngcr;
    CLEARBUF

    if ( writeserver(cl, "HEAD %s\r\n", msgid) == false ) return CHECK_ARTICLE_FAIL;

    if ( (readserverline(client->bbuf, cfg.BufSize)) == NULL ) return CHECK_ARTICLE_FAIL;
    if ( atoi(client->bbuf) > 399 ) i=0;

    checkbuf_init(chkbuf);
    while (i && ! client->error) {
        if ( (bread = nreadserver(client, client->bbuf, cfg.BufSize)) == 0 )
            return CHECK_ARTICLE_FAIL;

        if ( checkbuf_isend(chkbuf, client->bbuf, bread) ) 
            i = 0;

        if ( strstr(client->bbuf, "Newsgroups:") != NULL ) 
        {
            ng = strstr(client->bbuf, "Newsgroups:") + 12;
            for (ngcr=ng; (*ngcr) != '\r'; ngcr++) {};
            *ngcr=0;

	    /* read rest of head */
	    while ( i && ! client->error ) {
		bread = nreadserver(client, client->bbuf, cfg.BufSize);
        	if ( checkbuf_isend(chkbuf, client->bbuf, bread) ) 
		    i = 0;
	    }

            /* now we've found the article and have the newsgroups header */
            if ( match_expression(ng, getwildmat(cl->profile->ReadPat), 0) ) 
                return CHECK_ARTICLE_OK;
            else
                return CHECK_ARTICLE_NOPERM;
        }

        client->groupbytes  += (ulong)bread;
    }
    return CHECK_ARTICLE_FAIL;
}

/*
 * Basically a wrapper around find_article_server but
 * return some descriptive message to the client if either
 * the article is not found or we dont have permission
 * to retrieve it
 */
PROTO bool catch_art_server(CLIENT *cl, char *msgid) {
    int x;

    if ( cl->group == NULL )
    {
        if ( msgid[0] != '<' ) {
            swriteclient(cl, MSG_ARTICLECANT);
            return false;
        }

        if ( client->profile->ArticleSearch == 0 ) {
            swriteclient(cl, MSG_ARTICLENOPERM);
            return false;
        }

        x = find_article_server(cl, msgid);

        switch (x)
        {
            case CHECK_ARTICLE_FAIL:
                swriteclient(cl, MSG_ARTICLECANT);
                return false;

            case CHECK_ARTICLE_NOPERM:
                swriteclient(cl, MSG_ARTICLENOPERM);
                return false;

            case CHECK_ARTICLE_OK:
            default:
                return true;
        }
    }

    return true;
}

