/*
 * NNTPSwitch SYSV Semaphore functions
 *
 * No errorhandling on lock/unlock and close functions. If for some reason
 * the semaphore has gone we won't care, because the master process
 * has lost it counters anyway.
 *
 * $Id: semaphore.c,v 1.1 2004/11/11 13:32:02 tommy Exp $
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>

#include "aconfig.h"
#include "nntpd.h"
#include "aprotos.h"

key_t SEMKEY = 3333L;                      /* key to pass to semget() */

PROTO int semlock_init() 
{
    int i;

    if ((i = semget(SEMKEY, 1, IPC_CREAT | 0666)) == -1) 
    {
        die("semlock_init: semget failed %m");
    }
    return i;
}

PROTO void semlock_lock(int semid)
{
    struct sembuf sops[2];

    sops[0].sem_num = 0;                        /* We only use one track */
    sops[0].sem_op = 0;                         /* wait for semaphore flag to become zero */
    sops[0].sem_flg = SEM_UNDO;                 /* take off semaphore asynchronous  */

    sops[1].sem_num = 0;
    sops[1].sem_op = 1;                         /* increment semaphore -- take control of track */
    sops[1].sem_flg = SEM_UNDO;    		/* take off semaphore */

    semop(semid, sops, 2);
}
       		
PROTO void semlock_unlock(int semid) 
{
    struct sembuf sops[1];

    sops[0].sem_num = 0;
    sops[0].sem_op = -1;                        /* Give UP COntrol of track */
    sops[0].sem_flg = SEM_UNDO;    		/* take off semaphore, asynchronous  */

    semop(semid, sops, 1);
}


PROTO void semlock_close(int semid)
{
    semctl(semid, 0, IPC_RMID);
}

