/*
 * socket.c  Lowlevel socket server/client routines
 * 
 * $Id: socket.c,v 1.33 2004/11/21 01:11:49 tommy Exp $
 */

#include "aconfig.h"
#include "nntpd.h"
#include "aprotos.h"

#define CLEARBUFS client->sbuf[0]='\0';

#undef RL_DEBUG 


PROTO void init_ratelimiter_prof(CLIENT *cl) 
{
	cl->RateIntervalUs  = 1000000 / cfg.RatePeriod;
	gettimeofday(&cl->profile->rl_start, NULL);
	cl->profile->rl_sleep = RL_INITVAL;
}


PROTO int check_ratelimits_prof(CLIENT *cl) 
{
	struct timeval 	sleep, now;
	unsigned long 	timetaken, bytesdone;
	double 		bps;

	PROFILE *profile = cl->profile;

	if ( cl->profile->MaxProfileBPS == 0 ) 
		return 0;

	gettimeofday(&now, NULL);
	timersub(&now, &profile->rl_start, &sleep);
	timetaken = (sleep.tv_sec * 1000000) + sleep.tv_usec;

	if ( timetaken > cl->RateIntervalUs ) 
	{
		gettimeofday(&profile->rl_start, NULL);

        	bytesdone = profile->rl_curbytes - profile->rl_lastbytes;
		profile->rl_lastbytes = profile->rl_curbytes;

		/* calculate a reasonable average bps */
		bps = ((double)1000000 / timetaken) * bytesdone; 
		profile->rl_curbps = ((profile->rl_curbps * 2) + bps ) / 3;

		/*
	 	 * Now just flip/flop the delay. If we're going too fast,
	 	 * sleep 1.5 times more. If we're going too slow, sleep 1.5 
	 	 * times less. Eventually this value will tune itself.
	 	 */
		if ( profile->rl_curbps > cl->profile->MaxProfileBPS ) 
		{
			profile->rl_sleep *= 1.4;
#ifdef RL_DEBUG
			syslog(LOG_DEBUG, "sleep cl - %lf, bps = %lf, tt = %lu, bd = %lu", 
				user->rl_sleep, user->rl_curbps, timetaken, bytesdone);
#endif
		}else{
			profile->rl_sleep /= 1.4;
#ifdef RL_DEBUG
			syslog(LOG_DEBUG, "sleep cl + %lf, bps = %lf, tt = %lu, bd = %lu", 
				user->rl_sleep, user->rl_curbps, timetaken, bytesdone);
#endif
		}
	}

	/*
	 * Finally do the sleeping part, overruns can occur when unaccounted
	 * data is sent to the client. Underruns can occur when the process is often
	 * idle for example when reading text groups.
	*/

	if ( profile->rl_sleep > 3000000 ) 
	{
		profile->rl_sleep = 1000000;
	}else if ( profile->rl_sleep < 4000 ) 
	{
		profile->rl_sleep = RL_INITVAL;
	}
	usleep(profile->rl_sleep);

	return 0;
}


PROTO void init_ratelimiter(CLIENT *cl) 
{
	cl->RateIntervalUs  = 1000000 / cfg.RatePeriod;
	gettimeofday(&cl->user->rl_start, NULL);
	cl->user->rl_sleep = RL_INITVAL;
}


PROTO int check_ratelimits(CLIENT *cl) 
{
	struct timeval 	sleep, now;
	unsigned long 	timetaken, bytesdone;
	double 		bps;
	USER *user    = cl->user;

	if ( cl->profile->MaxUserBPS == 0 ) 
		return 0;

	gettimeofday(&now, NULL);
	timersub(&now, &user->rl_start, &sleep);
	timetaken = (sleep.tv_sec * 1000000) + sleep.tv_usec;

	if ( timetaken > cl->RateIntervalUs ) 
	{
		gettimeofday(&user->rl_start, NULL);

        	bytesdone = user->rl_curbytes - user->rl_lastbytes;
		user->rl_lastbytes = user->rl_curbytes;

		/* calculate a reasonable average bps */
		bps = ((double)1000000 / timetaken) * bytesdone; 
		user->rl_curbps = ((user->rl_curbps * 2) + bps ) / 3;

		/*
	 	 * Now just flip/flop the delay. If we're going too fast,
	 	 * sleep 1.5 times more. If we're going too slow, sleep 1.5 
	 	 * times less. Eventually this value will tune itself.
	 	 */
		if ( user->rl_curbps > cl->profile->MaxUserBPS ) 
		{
			user->rl_sleep *= 1.4;
#ifdef RL_DEBUG
			syslog(LOG_DEBUG, "sleep cl - %lf, bps = %lf, tt = %lu, bd = %lu", 
				user->rl_sleep, user->rl_curbps, timetaken, bytesdone);
#endif
		}else{
			user->rl_sleep /= 1.4;
#ifdef RL_DEBUG
			syslog(LOG_DEBUG, "sleep cl + %lf, bps = %lf, tt = %lu, bd = %lu", 
				user->rl_sleep, user->rl_curbps, timetaken, bytesdone);
#endif
		}
	}

	/*
	 * Finally do the sleeping part, overruns can occur when unaccounted
	 * data is sent to the client. Underruns can occur when the process is often
	 * idle for example when reading text groups.
	*/

	if ( user->rl_sleep > 3000000 ) 
	{
		user->rl_sleep = 1000000;
	}else if ( user->rl_sleep < 4000 ) 
	{
		user->rl_sleep = RL_INITVAL;
	}
	usleep(user->rl_sleep);

	return 0;
}


PROTO void set_errormsg(char *str, ...) 
{
	va_list ap;

	if ( client->error > 0 ) 
    	{
		/* we probably had a broken pipe somewhere by now.. */
        	syslog(LOG_NOTICE,"set_errormsg called twice %s", str);
        	return;
    	}

    	if ( (client->errstr = calloc(1,512)) == NULL )
        	die("No memory for errstr");

    	client->error++;
    	va_start(ap, str);
    	vsprintf(client->errstr, str, ap);
    	va_end(ap);
}


PROTO ssize_t read_socket(int socket, char *buf, int size, int timeout) 
{
	fd_set rdfs;
    	struct timeval tv;

    	FD_ZERO(&rdfs);
    	FD_SET(socket, &rdfs);

    	tv.tv_sec  = timeout;
    	tv.tv_usec = 0;

    	if ( select(socket+1, &rdfs, NULL, NULL, &tv) == 0 ) 
    	{
        	syslog(LOG_DEBUG, "read_socket: select timeout");
		return -1;
    	}

    	return read(socket, buf, size);
}


PROTO ssize_t write_socket(int socket, char *buf, int size, int timeout) 
{
	fd_set rdfs;
    	struct timeval tv;

    	FD_ZERO(&rdfs);
    	FD_SET(socket, &rdfs);

    	tv.tv_sec  = timeout;
    	tv.tv_usec = 0;

    	if ( select(socket+1, NULL, &rdfs, NULL, &tv) == 0 ) 
    	{
        	syslog(LOG_DEBUG, "write_socket: select timeout");
		return -1;
    	}

    	return write(socket, buf, size);
}


/* 
 * readserver() read a single line from the currently selected server
 */
PROTO char * readserver(CLIENT *client, char *buf, int size) 
{
    	fd_set rdfs;
    	struct timeval tv;
    	size_t bread;

    	FD_ZERO(&rdfs);
    	FD_SET(client->serversock, &rdfs);

    	tv.tv_sec  = cfg.ServerReadTimeout;
    	tv.tv_usec = 0;

    	if ( select(client->serversock+1, &rdfs, NULL, NULL, &tv) == 0 ) 
    	{
        	set_errormsg("remote server %s read timeout", client->lastserver);
		/* FIXME: Somethimes this has to be written to the client, sometimes not
	 	 * writeclient(client,"500 Remote server read timeout");
	 	 */
        	client->connected = 0;
		return NULL;
    	}

    	bread = read(client->serversock, buf, size);
    	if ( bread == -1 ) 
    	{
		set_errormsg("cant read from server %s %m", client->lastserver);
		bread=0;
    	}
    	buf[bread] = 0;

    	if ( cfg.LogReadserver )
        	syslog(LOG_DEBUG,"readserver: %s", buf);

    	client->serverbytes += (ulong)bread;
    	return buf;
}


/* 
 * nreadserver() same as readserver but return number of bytes read (or 0)
 */
PROTO int nreadserver(CLIENT *client, char *buf, int size) 
{
    	fd_set rdfs;
    	struct timeval tv;
    	size_t bread;

    	FD_ZERO(&rdfs);
    	FD_SET(client->serversock, &rdfs);

    	tv.tv_sec  = cfg.ServerReadTimeout;
    	tv.tv_usec = 0;

    	if ( select(client->serversock+1, &rdfs, NULL, NULL, &tv) == 0 ) 
    	{
        	set_errormsg("remote server %s read timeout", client->lastserver);
		/* FIXME
	 	 * writeclient(client,"500 Remote server read timeout");
	 	 */
        	client->connected = 0;
		return 0;
    	}

    	bread = read(client->serversock, buf, size);
    	if ( bread == -1 ) 
    	{
		set_errormsg("cant read from server %s %m", client->lastserver);
		return 0;
    	}
    	buf[bread] = 0;

    	if ( cfg.LogReadserver )
        	syslog(LOG_DEBUG,"readserver: %s", buf);

    	client->serverbytes += (ulong)bread;
    	return bread;
}


PROTO char * readserverline(char *buf, int size) 
{
    	fd_set rdfs;
    	struct timeval tv;
                                                                                                                           
    	FD_ZERO(&rdfs);
    	FD_SET(client->serversock, &rdfs);
                                                                                                                           
    	tv.tv_sec  = cfg.ServerReadTimeout;
    	tv.tv_usec = 0;
                                                                                                                           
    	if ( select(client->serversock+1, &rdfs, NULL, NULL, &tv) == 0 ) 
    	{
		set_errormsg("remote server %s read timeout", client->lastserver);
		/* FIXME
		 * writeclient(client,"500 Remote server read timeout");
		 */
        	client->connected = 0;
        	return NULL;
    	}
                                                                                                                           
    	return fgetsrn(buf,size,client->serversock);
}


/* 
 * readclient() read a single line from the client
 */
PROTO char * readclient(CLIENT *client, char *buf, int size, int nowait) 
{
    	fd_set rdfs;
    	struct timeval tv;

    	if ( ! nowait ) 
	{
		FD_ZERO(&rdfs);
        	FD_SET(client->socket, &rdfs);

        	tv.tv_sec  = cfg.ClientReadTimeout;
        	tv.tv_usec = 0;

		if ( select(client->socket+1, &rdfs, NULL, NULL, &tv) == 0 ) 
		{
	    		client->timeout++;
	    		return NULL;
		}
    	}
    	return fgetsrn(buf,size,client->socket);
}


/* 
 * nreadclient() same as readclient but returns num bytes read
 */
PROTO int nreadclient(CLIENT *client, char *buf, int size, int nowait) 
{
	fd_set rdfs;
    	struct timeval tv;

    	if ( ! nowait ) 
    	{
		FD_ZERO(&rdfs);
        	FD_SET(client->socket, &rdfs);

        	tv.tv_sec  = cfg.ClientReadTimeout;
        	tv.tv_usec = 0;

		if ( select(client->socket+1, &rdfs, NULL, NULL, &tv) == 0 ) 
		{
	    		client->timeout++;
	    		return 0;
		}
    	}

    	return nfgetsrn(buf,size,client->socket);
}


/*
 * writeserver() write to the currently selected server
 */
PROTO int writeserver(CLIENT *client, char *str, ...) 
{
	va_list ap;
    	size_t nc;
    	char *s;
    
    	va_start(ap, str);
    	CLEARBUFS
    	vsprintf(client->sbuf, str, ap);
    	s = fixrn(client->sbuf);
    
	if ( cfg.LogWriteserver ) 
		syslog(LOG_DEBUG,"writeserver %s",s);

    	if ( (nc = write(client->serversock, s, strlen(s))) == -1 ) 
    	{
        	syslog(LOG_ERR,"cant write to server %s %m", client->lastserver);
		return 0;
    	}

    	free(s);
    	va_end(ap);
    	return nc;
}


/*
 * writeclient() write to the client
 * return 0 on success, -1 on failure
 */
PROTO int writeclient(CLIENT *client, char *str, ...) 
{
    	va_list ap;
    	size_t nc;
    	char *s;
    	int n;
    
    	va_start(ap, str);
    	CLEARBUFS
    	vsprintf(client->sbuf, str, ap);
    	s = fixrn(client->sbuf);
    	va_end(ap);

    	if ( cfg.LogWriteclient && atoi(s) > 199 )
		syslog(LOG_DEBUG,"writeclient %s",s);

	n = strlen(s);
    	if ( (nc = write(client->socket, s, n)) == -1 ) 
    	{
		set_errormsg("cant write to client %m");
		return -1;
    	}
    	free(s);

    	return 0;
}


/*
 * swriteclient() statically write to the client
 * return 0 on success, -1 on failure
 * no vargs on this one
 */
PROTO int swriteclient(CLIENT *client, char *str) 
{
	char *s = fixrn(str);
    	int n;

    	if ( cfg.LogWriteclient && atoi(s) > 199 )
		syslog(LOG_DEBUG,"swriteclient %s",s);

    	n = strlen(s);
    	if ( (write(client->socket, s, n)) == -1 ) 
    	{
		set_errormsg("cant swrite to client %m");
		return -1;
    	}
    	free(s);

    	return 0;
}


/*
 * slwriteclient() statically write to the client and ratelimit
 * return 0 on success, -1 on failure
 * no vargs on this one
 */
PROTO int slwriteclient(CLIENT *client, char *str, size_t size, int limit) 
{
    	fd_set rdfs;
    	struct timeval tv;

    	if ( 1 ) 
    	{
		FD_ZERO(&rdfs);
        	FD_SET(client->socket, &rdfs);

        	tv.tv_sec  = cfg.ClientReadTimeout;
        	tv.tv_usec = 0;

		if ( select(client->socket+1, NULL, &rdfs, NULL, &tv) == 0 ) {
	    		set_errormsg("slwrite timeout");
	    		return -1;
		}
    	}
    
    	if ( (write(client->socket, str, size)) == -1 ) 
    	{
		set_errormsg("cant slwrite to client %m");
		return -1;
    	}

    	client->profile->rl_curbytes += size;
    	client->user->rl_curbytes +=size;
    	client->profile->bytes += size;
    
    	/* rate limit */
    	check_ratelimits(client);
    	check_ratelimits_prof(client);
    	return 0;
}


/**
 **
 **  Server Connection Functions
 **
 **/

PROTO int connect_socket(char *server, int port) 
{
    	struct sockaddr_in	name;
    	struct hostent	*hostinfo;
    	int sock;

    	syslog(LOG_NOTICE,"connecting to server %s",server);

    	sock = socket(PF_INET, SOCK_STREAM, 0);

    	name.sin_family = AF_INET;
    	if ( (hostinfo = gethostbyname(server)) == NULL ) 
    	{
		close(sock);
		return -1;
    	}

    	name.sin_addr = *((struct in_addr *)hostinfo->h_addr);
    	name.sin_port = htons(port);

    	if ( connect(sock, (struct sockaddr *) &name, sizeof(struct sockaddr)) == -1 ) 
	{
		syslog(LOG_ERR,"connect to server %s failed (%m)", server);
		close(sock);
		return -1;
    	}

    	return sock;
}


PROTO char* handshake_nntp(int sock, char *buf, int size, int timeout) 
{
    	int n;

    	if ( (n=read_socket(sock, buf, size, timeout)) <= 0 )
		return("Connection Failed: %m");

    	buf[n] = 0;

    	if ( atoi(buf) >= 300 )
		return buf;

    	return NULL;
}
		     

PROTO char* connect_auth_nntp(CLIENT* client, SERVER* server)
{
        writeserver(client, "AUTHINFO USER %s", server->Username);
        if ( readserver(client, client->sbuf, cfg.BufSize) == NULL ) 
                return MSG_AUTH_ERR;

        writeserver(client, "AUTHINFO PASS %s", server->Password);
        if ( readserver(client, client->sbuf, cfg.BufSize) == NULL ) 
                return MSG_AUTH_ERR;

        if ( atoi(client->sbuf) == 281 ) 
        {
                return NULL;
        }
        return client->sbuf;
}


PROTO int connect_server(CLIENT *client, SERVER *server)
{
        if ( connect_server_nntp(client, server->Hostname, server->Port, server->Timeout) ) 
        {
                return 1;
        }

        if ( server->Username[0] != 0 && server->Password[0] != 0 )
        {
                if ( connect_auth_nntp(client, server) != NULL ) 
                {
        	        syslog(LOG_ERR, "%s: remote server %s auth response %s", 
                                        client->hostname, server->Hostname, client->sbuf);
                        close(client->serversock);
                        client->connected=0;
                        client->lastserver[0] = 0;
		        return 1;
                }
        }

        return 0;
}


PROTO int connect_server_nntp(CLIENT *client, char *server, int port, int timeout) 
{
    	int sock;
    
    	if ( (sock=connect_socket(server, port)) == -1 )
		return 1;

    	if ( handshake_nntp(sock, client->sbuf, cfg.BufSize, timeout) != NULL ) 
    	{
        	syslog(LOG_ERR, "%s: error remote server %s response %s", client->hostname, server, client->sbuf);
        	close(sock);
		return 1;
	}

	client->serversock = sock;
    	client->connected=1;
    	client->serverarts=0;
    	client->serverbytes=0;
    	strncpy(client->lastserver, server, MAX_HOSTNAME-1);
    	return 0;
}


PROTO void send_quit(CLIENT *client) 
{
    	if ( client->connected == 1 ) 
    	{
		if ( ! writeserver(client, "QUIT") ) return;
        	if ( ! readserver(client, client->bbuf, cfg.BufSize) ) return;
    	}
}


/*
 * disconnect_server() disconnect from currently selected backend news server
 */
PROTO int disconnect_server(CLIENT *client) 
{
    	send_quit(client);
    	client->groupserver = NULL;
    	client->group = NULL;
    	close(client->serversock);
    	client->connected=0;
    	syslog(LOG_DEBUG, "disconnected from server %s articles %u bytes %llu", 
		client->lastserver, client->serverarts, client->serverbytes);
    	return 0;
}


PROTO char* fgetsrn(char* ptr, int maxlen, int fd)
{
	int	n, rc;
	char	c;

	for (n = 1; n < maxlen; n++) 
	{
		if ( (rc = read(fd, &c, 1)) == 1) 
		{
			*ptr++ = c;
			if (c == '\n')
				break;
		} else if (rc == 0) {
			if (n == 1)
				return(NULL);	/* EOF, no data read */
			else
				break;		/* EOF, some data was read */
		} else {
			if ( errno ) set_errormsg("fgetsrn: %m");
			return(NULL);	/* error */
		}
	}

	*ptr = 0;
	return(ptr);
}


PROTO int nfgetsrn(char* ptr, int maxlen, int fd)
{
	int	n, rc;
	char	c;

	for (n = 1; n < maxlen; n++) 
	{
		if ( (rc = read(fd, &c, 1)) == 1) 
		{
			*ptr++ = c;
			if (c == '\n')
				break;
		} else if (rc == 0) {
			if (n == 1)
				return(0);	/* EOF, no data read */
			else
				break;		/* EOF, some data was read */
		} else {
			if ( errno ) set_errormsg("nfgetsrn: %m");
			return(0);	/* error */
		}
	}

	*ptr = 0;
	return(n);
}

